/**
 * Copyright (c) 2013-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.extension.autoloader;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.TimerTask;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.jbi.JBIException;

import org.apache.commons.io.IOUtils;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.extension.JBIDescriptorExtensionBuilder;
import org.ow2.petals.jbi.descriptor.extension.exception.NoComponentNameDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NoIdentificationDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NotDeployableServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.exception.NotServiceUnitException;
import org.ow2.petals.jbi.descriptor.extension.generated.Identification;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Target;
import org.ow2.petals.microkernel.api.container.ComponentLifeCycle;
import org.ow2.petals.microkernel.api.container.Installer;
import org.ow2.petals.microkernel.api.jbi.management.DeploymentService;
import org.ow2.petals.microkernel.api.jbi.management.InstallationService;
import org.ow2.petals.microkernel.api.jbi.management.JbiArchiveType;
import org.ow2.petals.microkernel.api.jbi.management.XMLResult;
import org.ow2.petals.microkernel.api.util.NioUtil;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * The {@link TimerTask} scanning the installation directory
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class InstallDirectoryScanningTask extends TimerTask {

    private final static String TARGET_SU_ZIP_ARCHIVE = "service-unit.zip";

    private final DeploymentService deploymentService;

    private final InstallationService installationService;

    private final File installDir;

    private final ArchiveFileFilter archiveFilter = ArchiveFileFilter.getInstance();

    private final List<File> previousArchives = new ArrayList<File>();

    private final LoggingUtil logger;

    public InstallDirectoryScanningTask(final DeploymentService deploymentService,
            final InstallationService installationService, final File installDir, final LoggingUtil logger) {
        this.deploymentService = deploymentService;
        this.installationService = installationService;
        this.installDir = installDir;
        this.logger = logger;
    }

    @Override
    public void run() {
        final File[] newInstalls = this.installDir.listFiles(this.archiveFilter);
        if (newInstalls != null) {
            final List<File> newInstallList = new ArrayList<File>();

            synchronized (this.previousArchives) {

                // get only new archives
                for (final File archive : newInstalls) {
                    if (!this.previousArchives.contains(archive)) {
                        newInstallList.add(archive);
                    }
                }
                this.previousArchives.clear();

                // remove archives that are not completely copied in the install
                // directory
                final List<File> deleteInstallList = new ArrayList<File>();
                for (final File newArchive : newInstallList) {
                    if (!NioUtil.checkAccess(newArchive)) {
                        deleteInstallList.add(newArchive);
                    }
                }
                newInstallList.removeAll(deleteInstallList);

                if (!newInstallList.isEmpty()) {
                    this.install(newInstallList);
                    this.previousArchives.addAll(newInstallList);
                }
            }
        } else {
            this.logger.warning("Error when the autoloader scanned the directory '"
                    + this.installDir.getPath() + "'. Perhaps, the directory was deleted !");
        }
    }

    /**
     * Install the given list of files.
     * 
     * @param archiveFiles
     *            the list of JBI archives
     */
    private void install(final List<File> archiveFiles) {
        this.logger.start();

        final List<File> slToInstall = this.getJBIArchives(JbiArchiveType.SL_TYPE, archiveFiles);
        for (final File archive : slToInstall) {
            try {
                this.performInstallSL(archive);
            } catch (final Throwable e) {
                this.logger.error("Failed to auto-install of a shared library", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(slToInstall);

        final List<File> compToInstall = this.getJBIArchives(JbiArchiveType.COMPONENT_TYPE,
                archiveFiles);
        for (final File archive : compToInstall) {
            try {
                this.performInstallComponent(archive);
            } catch (final Exception e) {
                this.logger.error("Error during the auto-installation of a component", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(compToInstall);

        final List<File> saToInstall = this.getJBIArchives(JbiArchiveType.SA_TYPE, archiveFiles);
        for (final File archive : saToInstall) {
            try {
                this.performDeploySA(archive);
            } catch (final Throwable e) {
                this.logger.error("Error during the deployment of a service assembly: ", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(saToInstall);

        // Deployable service units
        final Iterator<File> itArchiveFiles = archiveFiles.iterator();
        while (itArchiveFiles.hasNext()) {
            final File archive = itArchiveFiles.next();
            try {
                final Jbi jbi = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptorFromArchive(archive);
                try {
                    try {
                        final Identification suIdent = JBIDescriptorExtensionBuilder.getInstance()
                                .getDeployableServiceUnitIndentification(jbi);
                        final String targetComponent = JBIDescriptorExtensionBuilder.getInstance()
                                .getDeployableServiceUnitTargetComponent(jbi);

                        try {
                            final File saFile = this.createSatoDeploySu(suIdent, targetComponent, archive.toURI()
                                    .toURL());
                            try {
                                this.performDeploySA(saFile);
                            } catch (final Exception e) {
                                this.logger.error("Error during the deployment of the temporary service assembly: ", e);
                            } finally {
                                try {
                                    if (!saFile.delete()) {
                                        this.logger.warning("Can't delete " + saFile.getAbsolutePath());
                                    }
                                } catch (final SecurityException e) {
                                    this.logger.error(e.getMessage(), e);
                                }
                            }
                        } catch (final Exception e) {
                            this.logger.error("Can't create a temporary service assembly: ", e);
                        } finally {
                            try {
                                if (!archive.delete()) {
                                    this.logger.warning("Can't delete " + archive.getAbsolutePath());
                                }
                            } catch (final SecurityException e) {
                                this.logger.error(e.getMessage(), e);
                            }
                        }

                    } catch (final NoComponentNameDeployableServiceUnitException e) {
                        // The archive is not a complete deployable service unit
                        this.logger.warning("Uncomplete deployable service unit. Target component name missing.");
                    } catch (final NoIdentificationDeployableServiceUnitException e) {
                        // The archive is not a complete deployable service unit
                        this.logger.warning("Uncomplete deployable service unit. Identification missing.");
                    }

                    itArchiveFiles.remove();

                } catch (final NotDeployableServiceUnitException e) {
                    // The archive is not a deployable service unit
                } catch (final NotServiceUnitException e) {
                    // The archive is not a service unit
                }
            } catch (final JBIDescriptorException e) {
                this.logger.warning("Unable to read a JBI descriptor", e);
            }
        }

        // clean the remaining archives
        for (final File archive : archiveFiles) {
            try {
                this.logger.warning("The archive '" + archive.getName()
                        + "' was deleted by Petals Autoloader because it is not a supported archive.");
                archive.delete();
            } catch (final SecurityException e) {
                this.logger.error(e.getMessage(), e);
            }
        }

        this.logger.end();
    }

    /**
     * Install the specified shared library in the JBI container Any exception
     * will cause the abortion of the process. A file "fileName.success" will be
     * created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     * 
     * @param slFile
     *            The shared library
     */
    private void performInstallSL(File slFile) {
        this.logger.call();

        try {
            this.installationService.installSharedLibrary(slFile.toURI().toString());
        } catch (final RuntimeException e) {
            this.logger.error("Failed to auto-install a shared library", e);
        }
    }

    /**
     * Install the specified Archive in the JBI container and start it . Any exception will cause the abortion of the
     * process. A file "fileName.success" will be created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     * 
     * @param componentFile
     *            The component
     */
    private void performInstallComponent(File componentFile) throws MalformedURLException, JBIException {
        this.logger.call();
        final Installer newInstaller = this.installationService
                .loadNewInstallerInternal(componentFile.toURI().toURL().toExternalForm());
        final ComponentLifeCycle clc = newInstaller.installInternal();
        clc.start();
    }

    /**
     * Get the set of files to install according to the JBI artefact type.
     * 
     * @param jbiType
     *            The expected JBI artefact
     * @return The list of selected JBI artefact files
     * @throws IOException
     * @throws JBIDescriptorException
     * @throws MalformedURLException
     */
    private List<File> getJBIArchives(final JbiArchiveType jbiType, final List<File> archivesToUninstall) {
        final List<File> archives = new ArrayList<File>();

        for (final File archive : archivesToUninstall) {
            try {
                final Jbi descriptor = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptorFromArchive(archive);
                if (JbiArchiveType.SL_TYPE.equals(jbiType) && descriptor.getSharedLibrary() != null) {
                    archives.add(archive);
                } else if (JbiArchiveType.COMPONENT_TYPE.equals(jbiType)
                        && descriptor.getComponent() != null) {
                    archives.add(archive);
                } else if (JbiArchiveType.SA_TYPE.equals(jbiType)
                        && descriptor.getServiceAssembly() != null) {
                    archives.add(archive);
                }
            } catch (JBIDescriptorException e) {
                this.logger.warning("Unable to read a JBI descriptor", e);
            }
        }

        return archives;
    }

    /**
     * Deploy the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param saFile
     *            the archive file
     * @throws Exception
     */
    private void performDeploySA(File saFile) throws Exception {
        this.logger.call();

        final String xml = this.deploymentService.deploy(saFile.toURI().toString());

        if (xml.indexOf(XMLResult.TaskResult.FAILED.toString()) > -1) {
            throw new RuntimeException("Failed to auto-deploy a service assembly");
        } else {
            final String saName = xml.substring(xml.indexOf("<loc-param>") + 11,
                    xml.indexOf("</loc-param>"));
            this.deploymentService.start(saName);
        }
    }

    /**
     * Create a {@link ServiceAssembly} and its archive to deploy the deployable service unit provided.
     * 
     * @param suIdentification
     *            The identification part of the deployable service unit JBI descriptor
     * @param targetComponentName
     *            The target component part of the deployable service unit JBI descriptor
     * @param suUrl
     *            The archive of the provided service unit
     * @return The archive of the resulting service assembly
     */
    private final File createSatoDeploySu(final Identification suIdentification, final String targetComponentName,
            final URL suUrl) throws Exception {

        final Jbi saJbiDescr = new Jbi();
        saJbiDescr.setVersion(BigDecimal.valueOf(1));
        final org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly saJbiDescrSa = new org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly();
        saJbiDescr.setServiceAssembly(saJbiDescrSa);
        final org.ow2.petals.jbi.descriptor.original.generated.Identification saJbiDescrIdent = new org.ow2.petals.jbi.descriptor.original.generated.Identification();
        saJbiDescrSa.setIdentification(saJbiDescrIdent);
        saJbiDescrIdent.setName(suIdentification.getName());
        saJbiDescrIdent.setDescription("SA automatically generated to deploy SU '" + suIdentification.getName() + "'");
        final org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit saJbiDescrSu = new org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit();
        saJbiDescrSa.getServiceUnit().add(saJbiDescrSu);
        final org.ow2.petals.jbi.descriptor.original.generated.Identification saJbiDescrSuIdent = new org.ow2.petals.jbi.descriptor.original.generated.Identification();
        saJbiDescrSuIdent.setName(suIdentification.getName());
        saJbiDescrSuIdent.setDescription("Original SU deployed");
        saJbiDescrSu.setIdentification(saJbiDescrSuIdent);
        final Target saJbiDescrSuTarget = new Target();
        saJbiDescrSuTarget.setArtifactsZip(TARGET_SU_ZIP_ARCHIVE);
        saJbiDescrSuTarget.setComponentName(targetComponentName);
        saJbiDescrSu.setTarget(saJbiDescrSuTarget);

        final File saFile = File.createTempFile("sa-" + suIdentification.getName() + "-", ".zip");
        final ZipOutputStream saZos = new ZipOutputStream(new FileOutputStream(saFile));
        try {
            final ZipEntry jbiEntry = new ZipEntry(JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE_IN_ARCHIVE);
            saZos.putNextEntry(jbiEntry);
            try {
                JBIDescriptorBuilder.getInstance().writeXMLJBIdescriptor(saJbiDescr, saZos);
            } finally {
                saZos.closeEntry();
            }
            final ZipEntry suArchiveEntry = new ZipEntry(TARGET_SU_ZIP_ARCHIVE);
            saZos.putNextEntry(suArchiveEntry);
            try {
                final InputStream isServiceUnit = suUrl.openStream();
                IOUtils.copy(isServiceUnit, saZos);
            } finally {
                saZos.closeEntry();
            }
        } finally {
            try {
                saZos.close();
            } catch (final IOException e) {
                this.logger.error("Can't close zip stream", e);
            }
        }

        return saFile;
    }

}
